
## for parsing GenePattern inputs
suppressMessages(suppressWarnings(library(getopt)))
suppressMessages(suppressWarnings(library(optparse)))


## libraries needed by this module
suppressMessages (suppressWarnings (library (mixtools)))
suppressMessages (suppressWarnings (library (mclust)))
suppressMessages (suppressWarnings (library (ggplot2)))
suppressMessages (suppressWarnings (library (reshape)))


sessionInfo ()


two.comp.normalize <- function (sample, type) {
  #   1. For all sample types, fit a 2-component gaussian mixture model using normalmixEM.
  #   2. For the bimodal samples, find the major mode M1 by kernel density estimation
  #     2a. Fit the model with one component mean constrained to equal M1
  #     2b. Normalize (standardize) samples using mean (M1) and resulting std. dev.
  #   3. For unimodal samples, find the mode M using kernel density estimation
  #     3a. Fit the model with mean for both components constrained to be equal to M
  #     3b. Normalize (standardize) samples using mean M and smaller std. dev. from model fit
  
  # WARNING:
  # This code has a lot of hacks to fix the flakiness of normalmixEM, and the idiosyncracies
  # of the actual data. Carefully re-examine code for new or altered input data
  
  data <- sample [ !is.na (sample) ]
  dens <- density (data, kernel='gaussian', bw='SJ')     # gaussian kernel with S-J bandwidth
                                                         # (see Venalbles & Ripley, 2002, pg, 129)
  # find major (highest) mode > -3 (to avoid problems with lower mode having higher density than higher mode)
  x.range <- dens$x > -3  
  dens.x <- dens$x [x.range];  dens.y <- dens$y [x.range]
  mode <- dens.x[which.max(dens.y)]                                                        
  if (type=='bimodal') mean.constr <- c (NA, mode) else mean.constr <- c (mode, mode)
  model <- normalmixEM (data, k=2, mean.constr=mean.constr)
  model.rep <- normalmixEM (data, k=2, mean.constr=mean.constr)
  model.alt <- Mclust (data, G=2, modelNames="V")
  alt.mu <- model.alt$parameters$mean
  alt.sd <- sqrt (model.alt$parameters$variance$sigmasq)
  # find reproducible model fit that is close to Mclust fit
  # if not, re-fit model -- without this condition
  # normalmixEM produces one-off model fits 
  n.try <- 1
  if (type=='unimodal') model.mode <- which(model$mu==mode)[which.min (model$sigma)]
  else model.mode <- which(model$mu==mode)
  model.other <- model.mode %% 2 + 1
  alt.mode <- which.min(abs(model.alt$par$mean-mode)); alt.other <- alt.mode %% 2 + 1
  while ( abs (model$mu[model.mode] - alt.mu[alt.mode]) > 3e-1 || abs (model$sigma[model.mode]-alt.sd[alt.mode]) > 3e-1 ||
          model$sigma[model.mode] < 0.1 ||
          (type=='bimodal' && (abs (model$mu[model.other] - alt.mu[alt.other]) > 1e1)) ||
          !all (c (model$mu, model$sigma) - c (model.rep$mu, model.rep$sigma) < 1e-3) ) {
    # if major mode (and SD of mode) is not within 0.3, or if the other mean (for bimodals only) 
    # is not within 1 of the Mclust result, try again
    model <- normalmixEM (data, k=2, mean.constr=mean.constr)
    model.rep <- normalmixEM (data, k=2, mean.constr=mean.constr)
    
    if (n.try > 50) stop (paste ("Can't fit mixture model ... giving up\n"))
    n.try <- n.try + 1
  }


  if (type=='bimodal') {
    # sometimes (esp. in phosphoproteome) the minor (lower) mode can be larger than the major (higher) mode
    # this situation is not possible in the unimodal samples
    corrected.mode <- model$mu [which.max(model$mu)]
    if (corrected.mode != mode) {
      cat ('  Lower mode larger than higher mode\n')
      mode <- corrected.mode
    }
  }
  norm.mean <- mode
  norm.sd <- ifelse (type=='bimodal', model$sigma[which(model$mu==mode)], min (model$sigma))
  
  # normalize by standardizing
  data <- data - norm.mean
  data <- data / norm.sd
  
  # return normalized data reorganized to original order
  sample [ !is.na (sample) ] <- data
  return ( list (norm.sample=sample, norm.mean=norm.mean, norm.sd=norm.sd, fit=unlist (c(model$mu, model$sigma))) )
}



normalize.dataset <- function (input.file, file.prefix, cls.file) {
  d <- read.gct ( input.file, check.names=FALSE )
  s <- colnames (d)
  cls <- read.cls (cls.file)
  
  norm.params <- NULL
  all.fits <- NULL
  temp <- lapply (3:ncol(d), 
                  function (i) {
                    cat ('Sample', s[i], '\n')
                    res <- two.comp.normalize (d[,i], cls[i-2])
                    d[,i] <<- res$norm.sample
                    norm.params <<- rbind (norm.params, c (sample=s[i], type=cls[i-2], mean=res$norm.mean, sd=res$norm.sd))
                    all.fits <<- rbind (all.fits, c (sample=s[i], type=cls[i-2], res$fit))
                  })
  
  write.gct (d, paste (file.prefix, '-norm.gct', sep=''))

  # plot profiles to check possible issues
  ## check if ratio profiles are similar (plot)

  pdf ( paste (file.prefix, '-norm-profile-plot.pdf', sep=''), width=6, height=5, pointsize=8 )
  for (cl in c ('unimodal', 'bimodal')) {
    data.norm <- melt (d [, c (FALSE, FALSE, cls==cl)])
    if (ncol (data.norm) > 0) {
      g <- ggplot (aes (x=value, color=variable), data=data.norm) + geom_density() + xlab ('ratio') + theme (legend.position='none') + ggtitle (paste ('Sample type:', cl))
      print (g)
    }
  }
  dev.off ()

  
  return (list (norm.params=norm.params, fits=all.fits))                    
}



# Get the command line arguments.  We'll process these with optparse.
# https://cran.r-project.org/web/packages/optparse/index.html
arguments <- commandArgs(trailingOnly=TRUE)

# Declare an option list for optparse to use in parsing the command line.
option_list <- list(
  # Note: it's not necessary for the names to match here, it's just a convention
  # to keep things consistent.
  make_option("--libdir", dest="libdir"),
  make_option("--gct.file", dest="input.file"),
  make_option("--output.prefix", dest="output.prefix"),
  make_option("--data.type", dest="type", default="proteome"),
  make_option("--cls.file", dest="bimodal.cls", default=NULL)
)

# Parse the command line arguments with the option list, printing the result
# to give a record as with sessionInfo.
opt <- parse_args (OptionParser (option_list=option_list), positional_arguments=TRUE, args=arguments)
print (opt)
opts <- opt$options


## extract options into appropriate variable
type <- opts$type
gct.file <- opts$input.file
gct.prefix <- opts$output.prefix

if (is.null(opts$bimodal.cls)) {
  bimodal.cls <- 'bimodal.cls'
  n <- scan (gct.file, skip=1, what=0, nlines=1)[2]
  cat (n, ' 1 1\n# unimodal\n', paste (rep (0, n), collapse=' '), '\n', sep='', file=bimodal.cls)
} else bimodal.cls <- opts$bimodal.cls


## read in supporting functions
source (file.path (opts$libdir, "io.r"))


## Run Normalization and plot results
# Normalize and save normalization statistics
stats <- normalize.dataset (gct.file, gct.prefix, bimodal.cls)
colnames (stats$fits) <- c ('sample', 'type', 'mean1', 'mean2', 'sd1', 'sd2')
write.csv (stats, paste (gct.prefix, '-norm-stats.csv', sep=''), row.names=FALSE)


# Plot summary of normalization statistics
plot.data <- data.frame (stats$norm.params)
for (x in c('mean', 'sd')) plot.data[,x] <- as.numeric (as.character (plot.data[,x]))
pdf ( paste (gct.prefix, '-norm-stats.pdf', sep=''), width=6, height=4)
g1 <- ggplot (aes (y=mean, x=type, col=type), data=plot.data) + geom_boxplot() + ggtitle ('Normalization factor: Mean'); print (g1)
g2 <- ggplot (aes (y=sd, x=type, col=type), data=plot.data) + geom_boxplot() + ggtitle ('Normalization factor: Std. Dev.'); print (g2)
dev.off ()

